import asyncio
import datetime

from sanic import Sanic
from sanic.response import json, text
from sanic.exceptions import RequestTimeout, NotFound

from sanic_openapi import swagger_blueprint
from sanic_useragent import SanicUserAgent
from sanic_sentry import SanicSentry
from sanic_compress import Compress
from sanic_cors import CORS

from xintian import load_config
from xintian.utils import CustomHandler
from xintian.db import RedisPool, MySQLPool, MongoPool
from xintian.exception import ServerError

config = load_config()
appid = config.get('APP_ID', __name__)
app = Sanic(appid, error_handler=CustomHandler())
app.config = config

app.blueprint(swagger_blueprint)

Compress(app)
SanicUserAgent.init_app(app)
SanicSentry(app)
CORS(app, automatic_options=True, esources={r"/*": {"origins": "*"}})


@app.listener('before_server_start')
async def before_server_start(app, loop):
    queue = asyncio.Queue()
    app.queue = queue
    app.redis_pool = None
    app.mysql_pool = None
    app.mongo_pool = None

    # redis
    if 'REDIS_CONFIG' in app.config:
        redis_pool = RedisPool()
        await redis_pool.open(app.config['REDIS_CONFIG'])
        app.redis_pool = redis_pool

    # mysql
    if 'MYSQL_CONFIG' in app.config:
        mysql_pool = MySQLPool()
        await mysql_pool.open(app.config['MYSQL_CONFIG'])
        app.mysql_pool = mysql_pool

        # mongo
    if 'MONGO_CONFIG' in app.config:
        mongo_pool = MongoPool()
        app.mongo_pool = await mongo_pool.open(app.config['MONGO_CONFIG'])


@app.listener('after_server_start')
async def after_server_start(app, loop):
    pass


@app.listener('before_server_stop')
async def before_server_stop(app, loop):
    app.queue.join()
    if app.redis_pool is not None:
        app.redis_pool.close()
    if app.mongo_pool is not None:
        app.mongo_pool.close()
    if app.mysql_pool is not None:
        app.mysql_pool.close()
        await app.mysql_pool.wait_closed()


@app.exception(RequestTimeout)
def timeout(request, exception):
    return json({'message': 'Request Timeout'}, 408)


@app.exception(NotFound)
def notfound(request, exception):
    return json(
        {'message': 'Requested URL {} not found'.format(request.url)}, 404)


@app.route('/robots.txt')
def robots(request):
    return text('Ignore me.')


@app.route('/favicon.ico')
def favicon(request):
    return text('Ignore me.')


@app.route('/health')
async def health(request):
    """
    For Kubernetes liveness probe，
    """
    try:
        # check redis valid
        if app.redis_pool:
            await app.redis_pool.save('health', 'ok', 1)

        # check mysql valid
        if app.mysql_pool:
            sql = "SELECT 666"
            result = await app.mysql_pool.fetchone(sql)
            if result is None:
                raise ServerError(error='内部错误', code='10500', message="msg")

    except Exception as e:
        raise ServerError(error='内部错误', code='10500', message="msg")

    return json({
        'pong': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
        'version': app.config['API_VERSION']
    })


@app.route('/ready')
async def ready(request):
    """
    For Kubernetes readiness probe，
    """
    try:
        # check redis valid.
        if app.redis_pool:
            await app.redis_pool.save('health', 'ok', 1)

        # check mysql valid.
        if app.mysql_pool:
            sql = "SELECT 666"
            result = await app.mysql_pool.fetchone(sql)
            if result is None:
                raise ServerError(error='内部错误', code='10500', message="msg")
    except Exception as e:
        raise ServerError(error='内部错误', code='10500', message="msg")

    return json({
        'pong': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
        'version': app.config['API_VERSION']
    })



import asyncio_redis
import aiomysql
from motor.motor_asyncio import AsyncIOMotorClient


class RedisPool:
    _pool = None
    redis_dict = None

    async def open(self, conf):
        if not self._pool:
            self.redis_dict = conf
            self._pool = await asyncio_redis.Pool.create(
                host=self.redis_dict['host'],
                port=self.redis_dict['port'],
                poolsize=self.redis_dict['pool_size'],
                password=None if self.redis_dict['password'] in ('', None) else self.redis_dict['password'],
                db=self.redis_dict['db']
            )

        return self._pool

    async def save(self, key, value, expire=15):
        redis_dict = self.redis_dict
        return await self._pool.set(key=redis_dict['namespace'] + ':' + key, value=value, expire=expire)

    async def load(self, key):
        redis_dict = self.redis_dict
        return await self._pool.get(redis_dict['namespace'] + ':' + key)

    async def close(self):
        self._pool.close()


class MySQLPool:
    _pool = None
    mysql_dict = None

    async def open(self, conf):
        if not self._pool:
            self.mysql_dict = conf
            self._pool = await aiomysql.create_pool(
                host=self.mysql_dict['host'],
                port=self.mysql_dict['port'],
                user=self.mysql_dict['user'],
                password=self.mysql_dict['password'],
                db=self.mysql_dict['db'],
                maxsize=5,
                autocommit=True,
                use_unicode=True,
                charset="utf8"
            )

        return self._pool

    async def fetchone(self, query: str, *args):
        async with self._pool.acquire() as conn:
            async with conn.cursor(aiomysql.DictCursor) as cur:
                await cur.execute(query, *args)
                return await cur.fetchone()

    async def fetchmany(self, query: str, *args):
        async with self._pool.acquire() as conn:
            async with conn.cursor(aiomysql.DictCursor) as cur:
                await cur.execute(query, *args)
                return await cur.fetchmany()

    async def fetchall(self, query: str, *args):
        async with self._pool.acquire() as conn:
            async with conn.cursor(aiomysql.DictCursor) as cur:
                await cur.execute(query, *args)
                return await cur.fetchall()

    async def close(self):
        self._pool.close()


class MongoPool:
    _pool = None

    async def open(self, conf):
        if not self._pool:
            self.mongo_dict = conf
            mongo_uri = "mongodb://" + self.mongo_dict['host'] + ":" + str(self.mongo_dict['port']) + "/" + \
                        self.mongo_dict['db']
            self._pool = AsyncIOMotorClient(mongo_uri)[self.mongo_dict['db']]

        return self._pool

    async def close(self):
        self._pool.close()



import logging
import json

from sanic.handlers import ErrorHandler
from xintian.exception import CustomException

logger = logging.getLogger('xintian')


def jsonify(records):
    """
    Parse database record response into JSON format
    """
    return [dict(r.items()) for r in records]


class CustomHandler(ErrorHandler):

    def default(self, request, exception):
        if isinstance(exception, CustomException):
            data = {
                'message': exception.message,
                'code': exception.code,
            }
            if exception.error:
                data.update({'error': exception.error})
            return json.dumps(data, status=exception.status_code)
        return super().default(request, exception)



import os
from sanic.config import Config


def load_config():
    conf = Config()
    module = os.environ.get('SANIC_SETTINGS_MODULE', 'settings')
    if module:
        path = '{}.py'.format(module.replace('.', '/'))
        conf.from_pyfile(path)
    return conf



class CustomException(Exception):
    status_code = None
    code = 100001
    msg = None
    error = None

    def __int__(self, error=None, code=None, msg=None, status_code=None):
        super().__init__(msg)
        if msg:
            self.msg = msg
        if code:
            self.code = code
        if status_code:
            self.status_code = status_code
        self.error = error


class BadRequest(CustomException):
    status_code = 400
    message = "Bad Request"


class Unauthorized(CustomException):
    status_code = 401
    message = "Unauthorized"


class Forbidden(CustomException):
    status_code = 403
    message = "Not Found"


class NotFound(CustomException):
    status_code = 404
    message = "Not Found"


class NotAcceptable(CustomException):
    status_code = 406
    message = "Unauthorized"


class Gone(CustomException):
    status_code = 410
    message = "Unauthorized"


class Enhance(CustomException):
    status_code = 420
    message = "Enhance Your Calm"


class UnprocessableEntity(CustomException):
    status_code = 422
    message = "Unprocessable Entity"


class TooManyRequests(CustomException):
    status_code = 429
    message = "Too Many Requests"


class ServerError(CustomException):
    status_code = 500
    message = "Internal Server Error"


class BadGateway(CustomException):
    status_code = 502
    message = "Bad Gateway"


class ServiceUnavailable(CustomException):
    status_code = 503
    message = "Service Unavailable"


class GatewayTimeout(CustomException):
    status_code = 504
    message = "Gateway timeout"



from setuptools import setup, find_packages

setup(
    name="xintian",
    version="0.0.5",
    author="huangxinping",
    author_email="o0402@outlook.com",
    description="A micro service framework for Sanic, so easy.",
    license="MIT",
    url="https://github.com/huangxinping/xintian",  # project home page, if any
    keywords = ['micro', 'service', 'sanic', 'flask'],
    packages=find_packages(),

    install_requires=[
        'sanic>=19.3.1',
        'sanic-openapi>=0.5.3',
        'sanic-useragent>=0.1.2',
        'sanic-sentry>=0.1.7',
        'sanic-compress>=0.1.1',
        'sanic-cors>=0.9.8',
        'asyncio-redis>=0.15.1',
        'aiomysql>=0.0.20',
        'motor>=2.0.0',
        'uvloop>=0.12.0',
        'aiohttp>=3.5.4',
        'pyyaml>=5.1',
    ],
    package_data={
        'xintian': ['*.py', '*.yml'],
    },
)
